#ifndef FUNCTIONS_DEBUG
#define FUNCTIONS_DEBUG

#include "error.h"


#if defined(__i386__) && !defined(NDEBUG)
#
  #if defined (__WATCOMC__)
  #
     void callDebugger (void);
    #pragma aux break_point = "int 3";
     void nop (void);
    #pragma aux nop = "nop";
  #
  #endif
  
  #if defined(__linux__)
  #
    #define callDebugger() __asm__ __volatile__ ("int $0x03\n") /* No Debugger Call */
    #define nop()          __asm__ __volatile__ ("nop\n")       /* No Nop           */
  #
  #endif

#
#else
#
  #define callDebugger()  do {} while (0)
  #define nop()           do {} while (0)
#
#endif


#ifndef NDEBUG
#
   // ADD THE LINE AND FILE WHERE IT OCCURRED
  #define debugClass()							\
  if(!this)										\
  {												\
	callDebugger();								\
	throw EFatalInstance();						\
  }

  #define debugError(boolean_test)				\
  if(boolean_test)								\
  {												\
    callDebugger();								\
    throw EFatalRange();						\
  }

   #define debugIf(boolean_test,messy)			\
   if(boolean_test)								\
   {											\
	  callDebugger();							\
	  throw EFatalMessy(messy);					\
   }

   #define debugExit(messy)						\
   do {											\
	 callDebugger();							\
	 throw EFatalMessy(messy);					\
   } while(0)
#
#else
#
   #define debugClass()                do {} while(0) /* CLASS TEST */
   #define debugError(boolean_test)    do {} while(0) /* FUNCTION ENTRY TEST */
   #define debugIf(boolean_test,messy) do {} while(0) /* FATAL_IF TEST */
   #define debugExit(messy)            do {} while(0) /* FATAL MESG */
#
#endif

#ifndef NLOG
#
   #define debug_print_log(messy)				\
   do {											\
      print_log(messy,__FILE__,__LINE__);		\
   } while(0)
#
#else
#
   #define debug_print_log(messy)      do {} while (0) /* INFO MESG */
#
#endif

#endif
