#ifndef EXCEPTION_ERROR
#define EXCEPTION_ERROR


// *****************
// **   PRIVATE   **
// *****************


#include "print.h"

//   ROOT error and exceptions 
// ********************************

class ExceptionError
{
 private:
  char* File;
  int   Line;
  char* Message;

  public:
    ExceptionError(const char* file, int line, const char* message)
    {
      File     = (char*)file;
      Line     = line;
	  if(!message)
		Message = "No message set";
	  else
		Message = (char*)message;
    }

	const char* getFile()    { return File;    };
	int         getLine()    { return Line;    };
	const char* getMessage() { return Message; };
};


//  Exception creation
// ********************

#define inheritance(name,parent)						\
class name : public parent								\
{														\
  public:												\
    name (const char* file, int line, const char* msg )	\
      : parent (file,line, msg) {};						\
} /* No semi-colon here, the semi-colon is to put after each inheritance macro */
  /* Otherwise it distrub Emacs indenting system !!                            */


//  Exceptions Classes creation
// *****************************

inheritance(EException ,ExceptionError);                    /* All Exeception */

inheritance(EMemoryException        ,EException);           /* Memory allocation failed  */ 
inheritance(EInstanciationException ,EMemoryException);     /* Function is not finished  */

inheritance(ESystemException        ,EException);           /* System Call Error         */
inheritance(EIOException            ,ESystemException);     /* IO Error                  */

inheritance(EDebugException         ,EException);           /* For various use           */
inheritance(EMessageException       ,EDebugException);      /* To throw a simple text    */


//  Error Classes creation
// ************************

inheritance(EError     ,ExceptionError);                    /* All Error */

inheritance(EParameterException     ,EError);
inheritance(EPrototypeException     ,EParameterException);  /* Invalid function Param    */
inheritance(EOutOfRangeException    ,EParameterException);  /* Values are out of range   */



// ****************
// **   PUBLIC   **
// **************** 

#define report(messy) throw MessageException(messy);

//  Exception creation Alias
// **************************

#define Exception(messy)              EException              (__FILE__,__LINE__,messy)

#define MemoryException(messy)        EMemoryException        (__FILE__,__LINE__,messy)
#define InstanciationException(messy) EInstanciationException (__FILE__,__LINE__,messy)

#define SystemException(messy)        ESystemException        (__FILE__,__LINE__,messy)
#define IOException(messy)            EIOException            (__FILE__,__LINE__,messy)

#define DebugException(messy)         EDebugException         (__FILE__,__LINE__,messy)
#define MessageException(messy)       EMessageException       (__FILE__,__LINE__,messy)

//  Error creation Alias
// **********************

#define Error(messy)  EError  (__FILE__,__LINE__,messy)

#define ParameterException(messy)     EParameterException     (__FILE__,__LINE__,messy)
#define PrototypeException(messy)     EPrototypeException     (__FILE__,__LINE__,messy)
#define OutOfRangeException(messy)    EOutOfRangeEException   (__FILE__,__LINE__,messy)

	 
#endif
