#include "ftp.h"

Ftp::Ftp (int bufferSize)
  : InOut(bufferSize)
#ifdef USE_UDP
  , SocketUDP(bufferSize)
#else
  , SocketTCP(bufferSize)
#endif
{
  Order      = new char[5];
  ValidOrder = new char[20*5+1];
  Data       = new char[bufferSize];

  ValidOrder[0] = '\0';
  strcat(ValidOrder,"NOOP "); // Does nothing
  strcat(ValidOrder,"CONN "); // Connection attempt
  strcat(ValidOrder,"PING "); // Ping-Pong game
  strcat(ValidOrder,"PONG ");
  strcat(ValidOrder,"DECN "); // Deconnection
  strcat(ValidOrder,"SYST "); // Half-Implemented
  strcat(ValidOrder,"ECHO "); // Output message
  strcat(ValidOrder,"USER "); // Set user name
  strcat(ValidOrder,"PASS "); // set user passwd
  strcat(ValidOrder,"PWD  "); // ftp pwd command
  strcat(ValidOrder,"CWD  "); // ftp cd command
  strcat(ValidOrder,"SIZE "); // ftp file size
  strcat(ValidOrder,"PORT "); // Server Host and Port indication
  strcat(ValidOrder,"LIST "); // ftp ls commmand
  strcat(ValidOrder,"RETR "); // ftp get command
}

Ftp::~Ftp ()
{
  delete[] ValidOrder;
  delete[] Order;
  delete[] Data;
}

unsigned int  Ftp::headerSize ()
{
  return 7;
}

bool Ftp::isOrder(const char* string)
{
  return (strstr(ValidOrder,string) == NULL);
}

int Ftp::getErrorCode (const char* string)
{
  return (string[0] - '0');
} 

const char* Ftp::getOrder (const char* string)
{
  strncpy(Order,string+2,4);
  Order[4] = '\0';
  return Order;
}

const char* Ftp::getData (const char* string)
{
  size_t len;

  len = strlen(string)-headerSize()-1;
  strncpy(Data,string+headerSize(),getBufferSize());
  while (len>=0 && (Data[len] == '\n' || Data[len] == '\r'))	
	  Data[len--] = '\0';	
  return Data;
}


const char* Ftp::readCommand(bool recursif)
{
  const char* string;
  string = readString();
  
  if(strlen(string)<headerSize())
	{
	  writeCommand("5 NOOP Order Too short '%s'",string);
	  if (recursif)	return readCommand();
	  else		    return string;
	}
  
  if(string[0] < '0' && string[0]>'5')
	{
	  writeCommand("5 NOOP Invalid Protocole Information '%c'",string[0]);
	  if (recursif)	return readCommand();
	  else		    return string;
	}
  
  if(isOrder(getOrder(string)))
	{
	  writeCommand("1 NOOP Unknown order '%s'",getOrder(string));
	  if (recursif)	return readCommand();
	  else		    return string;
	}

  if (
	  getErrorCode(string) == 1 ||
	  getErrorCode(string) == 2 ||
	  getErrorCode(string) == 3 ||
	  getErrorCode(string) == 4 ||
	  getErrorCode(string) == 5 ||
	  strcmp(getOrder(string),"ECHO") == 0
	  )	
	{
	  outputString("%s\n",getData(string));
	}

  if (strcmp(getOrder(string),"PING") == 0)
	{
	  writeCommand("0 PONG ");
	  readCommand(recursif);
	}  

  if (strcmp(getOrder(string),"DECN") == 0)
	{
	  closeConnection();
	  exit(0);
	}

  if (string[6] == '+')		 
	readCommand(recursif);
  
  return string;
}

ssize_t Ftp::writeCommand  (const char* string, ...)
{
  va_list ap;

  va_start(ap,string);
  vsprintf(OutBuffer,string,ap);
  va_end(ap);

  return writeString(OutBuffer);
}
