#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <arpa/telnet.h>

#include "ftpc.h"

Ftpc::Ftpc (char* hostname, int port) : Ftp()
{
  clientOf(hostname,port);
}


void Ftpc::run()
{
  bool finished;

  const char* readed;
  char  order[100];

  // Not Good server - simply abort
  readed = readString();
  if(strncmp(getOrder(readed),"CONN",4))
	{
	  closeConnection();
	  exit(0);
	}

  writeCommand("0 CONN Pleased to meet you FTP-Like server");
  readed = readCommand();

  finished = FALSE;
  while(!finished)
	{
	  outputString("ftp> ");
	  getnstr(order,99);

	  if(strncasecmp(order,"help ",5) == 0)
		printHelp(order+5);
	  
	  else if(strcasecmp(order,"help") == 0)
		printHelp(order+4);
	  
	  else if(strncasecmp(order,"trace ",6) == 0)
		trace(order+6);
	  
	  else if(strcasecmp(order,"trace") == 0)
		trace(order+5);

	  else if(strcasecmp(order,"ping") == 0)
		ping();

	  else if(strcasecmp(order,"user") == 0)
		outputString("Usage: user <user name>\n");

	  else if(strncasecmp(order,"user ",5) == 0)
		sendUser(order+5);

	  else if(strcasecmp(order,"pass") == 0)
		outputString("Usage: pass <user password>\n");

	  else if(strncasecmp(order,"pass ",5) == 0)
		sendPass(order+5);
	  
	  else if(strcasecmp(order,"pwd") == 0)
		getCurrentDir(order+3);

	  else if(strcasecmp(order,"sys") == 0)
		getSystem(order+3);

	  else if(strncasecmp(order,"cd ",3) == 0)
		changeDir(order+3);

	  else if(strcasecmp(order,"cd") == 0)
		getCurrentDir(order+2);
	  	  
	  else if(strncasecmp(order,"ls ",3) == 0)
		getList(order+3);
	  
	  else if(strcasecmp(order,"ls") == 0)
		getList(order+2);
	  
	  else if(strncasecmp(order,"get ",4) == 0)
		getFile(order+4);

	  else if(strcasecmp(order,"get") == 0)
		outputString("Usage : get <filename>");
	  
	  else if(strcasecmp(order,"quit") == 0)
		finished = TRUE;
	  
	  else if(strcasecmp(order,"exit") == 0)
		finished = TRUE;
	  
	  else if(strcasecmp(order,"leave") == 0)
		finished = TRUE;	  
	  
	  // local command
	  else if(strncasecmp(order,"!cd ",4) == 0)
		localChangeDir(order+4);

	  else if(strcasecmp(order,"!cd") == 0)
		outputString("Usage : !cd <remote directory>");

	  else if (strncasecmp(order,"!",1) == 0)
		  system(order+1);

	  // Sending direct order to server
	  else if (strcasecmp(order,"order ") == 0)
		  writeString(order+6);
		
	  else
		if(strlen(order)>0)
		  outputString("Command unknown\n");
	}

  deconnection();
  exit(0);
}

void Ftpc::deconnection()
{
  writeCommand("0 DECN Bye bye"); 
}


void Ftpc::printHelp(const char* option)
{
  if(*option == '\0')
	{
	  outputString("Help : List of supported command\n");
	  outputString("--------------------------------\n");
	  outputString("Local         : help trace !name\n");
	  outputString("Information   : trace pwd\n");
	  outputString("Communication : cd dir ls\n");
	  outputString("Quit          : quit leave exit\n");
	}
  else
	{
	  outputString("No help on command .. Too lazy\n");
	}
}

void Ftpc::trace(const char* readed)
{
  if(strncasecmp(readed," on",3) == 0)
	setVerbose(TRUE);

  if(strncasecmp(readed," off",4) == 0)
	setVerbose(FALSE);

  outputString("trace is %s\n",(getVerbose())?"on":"off");
}

void Ftpc::ping()
{
  const char* readed;  

  writeString("0 PING ");
  readed = readCommand(FALSE);

  if(strcmp(getOrder(readed),"PONG") == 0)
	outputString("Got the pong\n");
  else
	outputString("Got this stange thing : '%s'\n",readed);
}

void Ftpc::sendUser(const char* option)
{
  writeCommand("0 USER %s\n",option);
  readCommand();
}

void Ftpc::sendPass(const char* option)
{
  writeCommand("0 PASS %s\n",option);
  readCommand();
}
 
void Ftpc::getSystem(const char* option)
{
  writeCommand("0 SYST ");
  readCommand(FALSE);
}

void Ftpc::getCurrentDir(const char* option)
{
  writeCommand("0 PWD  %s",option);
  readCommand(FALSE);
}

void Ftpc::changeDir(const char* option)
{
  writeCommand("0 CWD  %s",option);
  readCommand(FALSE);
}

bool Ftpc::getSize(off_t* size)
{
  const char* readed;

  writeCommand("0 SIZE ");
  readed = readCommand(FALSE);

  if(getErrorCode(readed) == 0)
	{
	  *size = atoi(readed+headerSize());
	  return TRUE;
	}
  else
	{
	  return FALSE;
	}
}

bool Ftpc::sendPort (Socket* server)
{
  writeCommand("0 PORT %s,%d,%d",server->getHostIp(','),server->getPort()/256,server->getPort()&0xFF);
  return (getErrorCode(readCommand(FALSE)) == 0);
}

bool Ftpc::createServer(Socket*& server)
{
  // Creating the channel
  try
	{
	  if(getProtocole() == TCP)
		server = new SocketTCP(getBufferSize()); 
	  else
		server = new SocketUDP(getBufferSize());
	}
  catch(ESocketException)
	{
	  outputString("Can't create a new socket\n");
	  return FALSE;
	}

  try
	{
	  server->serverOn(0);
	}
  catch(ESocketException)
	{
 	  delete server;
	  outputString("Can't create a reception server\n");
	  return FALSE;
	}

  return TRUE;
}

bool Ftpc::waitOnServer(const char* command, Socket* server)
{
  bool returned;

  try
	{
	  writeCommand(command);
	  if(getErrorCode(readCommand(FALSE)) == 0)
		{
		  server->accept();
		  returned = TRUE;
		}
	  else
		returned = FALSE;
	}
  catch(ESocketException)
	{
	  outputString("Connection to client failed \n");
	  returned = FALSE;
	}

  return returned;
}


bool Ftpc::getData(const char* command, int file)
{
  Socket* recv;
  off_t size;

  if(!createServer(recv))
	return FALSE;

  if(!sendPort(recv))
	return FALSE;
  
  if(!waitOnServer(command,recv))
	{
	  delete recv;
	  return FALSE;
	}

  // Read Size or error server message
  if(!getSize(&size)) // Must be after LIST/RECV/ETC..
	{
	  delete recv;
	  return FALSE;
	}


  // Reading and outputing the result

  try
	{
	  off_t counter;
	  char  car;

	  for(counter=0;counter<size;counter++)
		{
		  car = recv->readChar();
		  ::write(file,&car,1);
		}
	}
  catch (ESocketException)
	{
	  // TCP End of transfert (close pipe by server)	  
	}

  writeCommand("0 NOOP Synchro etablished");
  readCommand();
  delete recv;

  return TRUE;
}




void Ftpc::getList(const char* option)
{
  char* command;

  try
	{
	  command = new char[strlen(option)+headerSize()+2];
	}
  catch(...)
	{
	  outputString("Memory allocation failed");
	  return;
	}

  strcpy(command,"0 LIST ");
  strcat(command,option);

  getData(command,1);
  delete command;
}



void Ftpc::getFile(const char* option)
{
  char* command;

  int exist = open(option,O_RDONLY);
  if (exist >= 0)
  	{
	  outputString("File '%s' already exist\n",option);
	  return;
	}

  int file = open(option,O_CREAT|O_EXCL|O_RDWR,S_IRUSR|S_IWUSR);
  if(file == -1)
	{
	  outputString("File '%s' can't be created\n",option);
	  return;
	}

  try
	{
	  command = new char[strlen(option)+headerSize()+2];
	}
  catch(...)
	{
	  outputString("Memory allocation failed");
	  return;
	}

  strcpy(command,"0 RETR ");
  strcat(command,option);

  if(!getData(command,file))
	{
	  unlink(option);
	}
 
  close  (file);
  delete command;
}


void Ftpc::localChangeDir(const char* param)
{
  if(chdir(param) != 0)
	outputString("cd: %s: No such file or directory (or Permission denied)\n",param);
}
