#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <sys/wait.h>
#include "ftpd.h"


Ftpd::Ftpd (int port) : Ftp()
{
  ReadyToReceived = FALSE;
  serverOn(port);
}

void Ftpd::run()
{
  for(;;)
	switch(connection())
	  {
	  case 0:
		connectClient();
		break;
		
	  default:
		int   status;
		waitpid(-1,&status,WNOHANG);
		break;
	  }			
}

void Ftpd::connectClient()
{  
  bool allowedToEnter;
  const char* readed;

  allowedToEnter = FALSE;
  
  // Send wlecome message and wait forever a correct answer
  presentation();

  //  Identification
#ifdef IDENTIFICATION
  while(!allowedToEnter)
	{
	  readed = readCommand();
	  if(strncmp(getOrder(readed),"USER",4) == 0)
		allowedToEnter = setUser(getData(readed));
	  else
		writeCommand("4 USER You must first specify a User Name");
   	}
#endif
/*
  writeCommand("0 ECHO+Hello %s",ClientName);
  writeCommand("0 ECHO+Welcome On this server");
  writeCommand("0 ECHO End Of message");
*/

  for(;;)
	{
	  readed = readCommand();

	       if(strcmp(getOrder(readed),"PWD ") == 0)
		sendCurrentDir(getData(readed));
	  else if(strcmp(getOrder(readed),"CWD ") == 0)
		changeDir(getData(readed));
	  else if(strcmp(getOrder(readed),"PORT") == 0)
		readPort(getData(readed));
	  else if(strcmp(getOrder(readed),"LIST") == 0)
		sendList(getData(readed));
	  else if(strcmp(getOrder(readed),"RETR") == 0)
		sendFile(getData(readed));
	  else if(strcmp(getOrder(readed),"SYST") == 0)
		sendSystem(getData(readed));
	  else if(strcmp(getOrder(readed),"USER") == 0)
		setUser(getData(readed));
	  else
		writeCommand("3 ECHO Server didnt implement '%s'",getOrder(readed));
	}
}

void Ftpd::presentation()
{
  bool agree;
  const char* readed;

  // Send Welcome message
  writeCommand("0 CONN %s FTP-like server. ready.",getLocalHostName());

  agree = FALSE;  
  do
	{
	  // Wait protocole used	  
	  readed = readString();

	  // Good Protocol
	  if(!strcmp(getOrder(readed),"CONN"))
		{		  
		  agree = TRUE;
		  writeCommand("0 ECHO Please Specify an user");
		}
	  // Not expected protocole
	  else
		{
		  writeCommand("1 NOOP Unknown Command");
		}		
	}
    while(!agree);
}

bool Ftpd::setUser (const char* readed)
{
  bool identified;
  char pass[20];
  
  identified = FALSE;	

  if(strlen(readed)>0)
	{
	  strncpy(UserName,readed,18);
	  writeCommand("0 ECHO Please specify User password");
	}
  else
	{
	  writeCommand("2 USER Command need a name");
	}

  readed = readCommand();

  if(strcmp(getOrder(readed),"PASS") == 0)
	{
	  if(strlen(getData(readed))>0)
		{
		  identified = TRUE;
		  strncpy(pass,getData(readed),18);
		  writeCommand("0 ECHO password successfull");
		}
	  else
		{
		  writeCommand("2 PASS The PASS command need an argument");
		}
	}
  else
	{
	  writeCommand("4 PASS You msut give the user password. Try to relog.");
	}
  
  if(identified)
	{	  
	  // Test User Name and Passwd
	  // su lui (if app is owned by root)
	}
  
  return identified;
}

void Ftpd::sendSystem(const char* param)
{
  writeCommand("0 ECHO Host System undefined");
}


void Ftpd::sendCurrentDir(const char* param)
{
  char directory[1024];

  getcwd(directory,1024);
  writeCommand("0 ECHO Current directory is %s",directory);
}

void Ftpd::changeDir(const char* param)
{
  if(chdir(param) == 0)
	writeCommand("0 ECHO directory changed");
  else
	writeCommand("2 CWD- Cant go in %s",param);
}


void Ftpd::readPort(const char* param)
{
  index_t nbVirgule;
  index_t nbNumber;
  index_t pos;
  index_t size;

  int high;
  int low;
  
  high = 0;
  low  = 0;

  nbVirgule = 0;
  nbNumber  = 0;
  pos       = 0;
  size      = strlen(param);

  while(pos<size && nbVirgule<4 && nbNumber <4)
	{
	  if(param[pos++] == ',')
		{
		  nbVirgule++;
		  nbNumber=0;
		}
	  else
		nbNumber++;
	}

  if(nbVirgule != 4)
	{
	  writeCommand("2 PORT %s",param);
	  return;
	}

  strncpy(ClientAddress,param,pos-1);
  ClientAddress[pos-1] = '\0';

  while(isdigit(param[pos]))
	{
	  high = high*10 + param[pos]-'0';
	  pos++;
	}
 
  if(param[pos++] != ',')
	{
	  writeCommand("2 PORT %s",param);
	  return;
	}

  while(isdigit(param[pos]))
	{
	  low = low*10 + param[pos]-'0';
	  pos++;
	}
  
  ClientPort = high*256 + low;

  pos=0;
  while(ClientAddress[pos] != '\0')
	{
	  if(ClientAddress[pos] == ',')
		ClientAddress[pos] = '.';
	  pos++;
	}

  ReadyToReceived = TRUE;
  writeCommand("0 PORT Command sucessfull");
}

bool Ftpd::sendSize (off_t size)
{
  const char* readed;
  readed = readCommand();
  
  if(strcmp(getOrder(readed),"SIZE") == 0)
	{
	  if(size != 0)
		{
		  writeCommand("0 SIZE %ld",size);
		}
	  else
		{
		  outputString("== File size is zero\n");
		  writeCommand("3 SIZE 0");
		  return FALSE;
		}	  
	}
  else
	{
	  outputString("== Order sequence waiting for SIZE received %s\n",getOrder(readed));
	  writeCommand("4 SIZE Server was waiting for a size\n");
	  return FALSE;
	}
  return TRUE;
}


void Ftpd::sendData(int file, off_t size)
{
  Socket* sender;
  char*   buffer;

  if(!ReadyToReceived)
	{
	  outputString("== No port defined can't send data\n");
	  writeCommand("4 ECHO You must specify a port first");
	  return;
	}

  try
	{
	  buffer  = new char[getBufferSize()];
	}
  catch(EMemoryException)
	{
	  writeCommand("3 NOOP Server memory allocation failed");
	  outputString("== Can't allocate memory\n");
	}

  try
	{
	  writeCommand("0 ECHO Connexion for transfert");
	  
	  if(getProtocole() == TCP) sender = new SocketTCP(getBufferSize());
	  else		                sender = new SocketUDP(getBufferSize());

	  sender->clientOf(ClientAddress,ClientPort);
	}
  catch(ESocketException)
	{
	  writeCommand("3 NOOP Server Connection failed");
	}
  catch(...)
	{
	  writeCommand("3 NOOP Server Can't create socket");
	  outputString("== Can't create a new socket\n");
	  return;
	}

  if(!sendSize (size)) return;
	  
  try
	{
	  off_t counter = 0;
	  off_t readed;
	  
	  outputString("== Tranfering Data to %s on Port %d\n",ClientAddress,ClientPort);

	  do
		{
		  readed = ::read(file,buffer,sender->getBufferSize());
		  if(readed > 0)
			sender->write(buffer,readed);
		  else				
			throw IOException("read failed");
		  counter += readed;
		}
	  while(counter < size);

	  outputString("== done.\n");
	}
  catch(EIOException E)
	{
	  writeCommand("3 NOOP Server I/O failed");
	}

  readCommand();
  writeCommand("0 ECHO Connection Closed");

  ReadyToReceived = FALSE;
  delete[] buffer;
  delete   sender;
}





void Ftpd::sendList(const char* param)
{
  FILE* file_tmp;
  int   file_out;
  int   copy_one;

  off_t file_size;

  char* command;  
  char* lsCommand = "/bin/ls ";
 
  try
	{
	  command = new char[strlen(param)+strlen(lsCommand)+1]; 
	}
  catch(EMemoryException)
	{
	  outputString("== Can't allocate memory\n");
	  writeCommand("3 LIST Server memory allocation failed");
	}  
  
  // Open temporary file
  
  file_tmp = tmpfile();
  file_out = fileno(file_tmp);
  
  if(!file_tmp)
	{
	  outputString("== Temp file creation failed\n");
	  writeCommand("3 LIST Temporary file creation failed");
	  return;
	}
  

  strcpy(command,lsCommand);
  strcat(command,param); // Is dangerous as LIST -al ; rm -rf ~ is valid

  // Get the command out put in the file

  copy_one = dup(1);
  dup2(file_out,1);

  system(command);

  dup2 (copy_one,1);
  close(copy_one);

  // Rewind the tempory file

  file_size = lseek(file_out,0,SEEK_CUR);
  if(file_size == -1)
	{
	  fclose(file_tmp);
	  outputString("== File seek failed\n");
	  writeCommand("3 LIST Can lseek on temporary file");
	  return;
	}

  if(lseek(file_out,0,SEEK_SET) != 0)
	{
	  fclose(file_tmp);
	  outputString("== File seek failed\n");
	  writeCommand("3 LIST Can lseek on temporary file");
	  return;
	}

  sendData(file_out,file_size);

  fclose(file_tmp);
  delete[] command;
}


void Ftpd::sendFile(const char* param)
{
  int file = open(param,O_RDONLY);
  if(file == -1)
	{
	  writeCommand("3 ECHO File '%s' can't be opened",param);
	  return;
	}

  struct stat info;
  if(stat(param,&info) != 0)
	{
	  writeCommand("3 ECHO File '%s' can't be asked for size",param);
	  return;
	}

  sendData(file,info.st_size);

  close(file);
}
