#include <stdlib.h>
#include <unistd.h>
#include <memory.h>
#include <ctype.h>
#include <signal.h>
//#include <setjmp.h>

#include "socket.h"

// Definition des Constantes

char* Socket::StringUDP = "udp";
char* Socket::StringTCP = "tcp";

Socket::Socket(int bufferSize)
{
  char hostName[255];

  LocalHostInfo  = NULL;
  RemoteHostInfo = NULL;

  SocketPort   = -1;
  SocketStream = -1;
  AcceptStream = -1;
  IOStream     = &SocketStream;
  Verbose      = FALSE;
	  
  if(gethostname(hostName,250))
	throw GetHostNameException("Unable to find local host name !!");	

  LocalHostInfo = gethostbyname(hostName);

  if(!LocalHostInfo)
	throw NoRouteToHostException("Unable to get local host Information");

  BufferSize = bufferSize;
  InBuffer  = new char[BufferSize];
  OutBuffer = new char[BufferSize];

  // ignore broken pipe errors
  signal(SIGPIPE,SIG_IGN);  
}

Socket::~Socket()
{
  delete[] InBuffer;
  delete[] OutBuffer;
  closeConnection();
}


void Socket::closeSocket()
{
  if(SocketStream != -1)
	{
	  close(SocketStream);
	  SocketStream = -1;
	}
}

void Socket::closeAccept()
{
  if(AcceptStream != -1)
	{
	  close(AcceptStream);
	  AcceptStream = -1;
	  IOStream = &SocketStream; 
	}
}

void Socket::closeConnection()
{
  closeSocket();
  closeAccept();
}

// Client Side

void Socket::clientOf(const char* hostname, port_t port)
{
  int returned;

  SocketPort = port;

  // Find Host on network

  RemoteHostInfo = gethostbyname(hostname);
  if(!RemoteHostInfo)
	throw NoRouteToHostException("Unable to find server host information");

  // Connection

  memset(&SocketInfo,0,sizeof(SocketInfo));

  SocketInfo.sin_family = AF_INET;
  memcpy
	( 
	 (char*)&(SocketInfo.sin_addr.s_addr),
	 RemoteHostInfo->h_addr_list[0],
	 RemoteHostInfo->h_length
	 );
  SocketInfo.sin_port = htons(port);
  
  returned = ::connect(SocketStream,(sockaddr*)&SocketInfo,sizeof(SocketInfo));
  
  if(returned < 0)
	throw ConnectException("Can't connect to socket");

  IOStream = &SocketStream;
}

// Server Side

void Socket::serverOn(port_t port)
{
  int returned;
  unsigned int sizeInfo;

  // Give a public name to socket

  memset(&SocketInfo,0,sizeof(SocketInfo));

  SocketInfo.sin_family      = LocalHostInfo->h_addrtype;
  SocketInfo.sin_addr.s_addr = INADDR_ANY;
  SocketInfo.sin_port        = htons(port);

  returned = bind(SocketStream,(struct sockaddr*)&SocketInfo,sizeof(SocketInfo));

  if(returned)
	throw ConnectException("Can't bind to socket");

  // Get Information back on the Socket

  sizeInfo = sizeof(SocketInfo);
  returned = getsockname(SocketStream,(struct sockaddr*)&SocketInfo,&sizeInfo);

  if(returned)
	throw SocketException("Can't get Socket Name");

  // The port MUST be read from SocketInfo as 0 -> auto-affecation
  SocketPort = ntohs(SocketInfo.sin_port);
}            

pid_t Socket::connection()
{
  pid_t pid;

  accept();

  pid = fork();
  switch(pid)
	{
	case -1:
	  throw SystemException("Can't fork");
	  break;
	  
	case 0:
	  closeSocket();
	  break;
	  
	default:
	  closeAccept();
	  break;
	}
  return pid;
}

// Commun Side



ssize_t Socket::read (char* buffer, ssize_t size)
{
  ssize_t number;

  number = 0;
  while (number<size)
	buffer[number++] = readChar();

  return number;
}

ssize_t Socket::write(const char* buffer, ssize_t size)
{
  ssize_t number;

  number = 0;
  while (number<size)
	writeChar(buffer[number++]);

  return number; 
}


const char* Socket::readString()
{
  ssize_t size = 0;
  do
	{
	  InBuffer[size] = readChar();
	}
  while(InBuffer[size] != '\0' && size++ < BufferSize-1);
  InBuffer[size] = '\0';

  printVerbose("-> '",InBuffer,"'");

  return InBuffer;
}

ssize_t Socket::writeText(const char* string)
{
  printVerbose("<- '",string,"'");
  return write(string,strlen(string)+1);
}

ssize_t Socket::writeString(const char* string, ...)
{
  va_list ap;

  va_start(ap,string);
  vsprintf(OutBuffer,string,ap);
  va_end(ap);

  return writeText(OutBuffer);
}


ssize_t Socket::skip (ssize_t size)
{
  return 0;
}



int Socket::getDescriptor()
{
  return *IOStream;
}

const char* Socket::getLocalHostName()
{
  return LocalHostInfo->h_name;
}

const char* Socket::getRemoteHostName()
{
  return RemoteHostInfo->h_name;
}


const char* Socket::getHostIp()
{
  return LocalHostInfo->h_addr_list[0];
}

const char* Socket::getHostIp(char separator)
{
  sprintf
	(HostIP,
	 "%u%c" "%u%c" "%u%c" "%u",
	 *((unsigned char*)(LocalHostInfo->h_addr_list[0]+0)),
	 separator,
	 *((unsigned char*)(LocalHostInfo->h_addr_list[0]+1)),
	 separator,
	 *((unsigned char*)(LocalHostInfo->h_addr_list[0]+2)),
	 separator,
	 *((unsigned char*)(LocalHostInfo->h_addr_list[0]+3))
	 );

  return HostIP;
}

index_t Socket::getBufferSize ()
{
  return BufferSize;
}

bool Socket::getVerbose ()
{
  return Verbose;
}

bool Socket::setVerbose (bool state)
{
  return (Verbose = state);
}

// Private

void Socket::printVerbose (const char* start, const char* string, const char* end)
{
  char letter;

  if(Verbose)
	{
	  printf("%s",start);
	  for(uint i=0; i<strlen(string);i++)
		{
		  letter = string[i];

		       if (letter == '\t')	printf("\\t");
		  else if (letter == '\r')	printf("\\r");
		  else if (letter == '\n')	printf("\\n");
		  else if (isprint(letter))	printf("%c",letter);
		  else			            printf("*");
		}
	  printf("%s\n",end);
	}
}

