#ifndef INCLUDE_NETWORK
#define INCLUDE_NETWORK

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>

#include "base.h"

// The new type used 

typedef int port_t;

typedef enum
{ 
  UDP = SOCK_DGRAM,
  TCP = SOCK_STREAM
} protocole_t;

// Creation of the Network exception

inheritance(ENetworkException        ,ESystemException);    // NetworkException
inheritance(EGetHostNameException    ,ESystemException);    // NoRouteToHostException
inheritance(ENoRouteToHostException  ,ENetworkException);   // NoRouteToHostException
inheritance(ESocketException         ,ENetworkException);   // SocketException
inheritance(EBindException           ,ESocketException);    // BindException
inheritance(EListenException         ,ESocketException);    // ListenException
inheritance(EConnectException        ,ESocketException);    // ConnectException
inheritance(EAcceptException         ,ESocketException);    // ConnectException
inheritance(ESocketAccessException   ,ESocketException);    // SocketAccessException

#define NetworkException(messy)       ENetworkException       (__FILE__,__LINE__,messy)
#define GetHostNameException(messy)   EGetHostNameException   (__FILE__,__LINE__,messy)
#define NoRouteToHostException(messy) ENoRouteToHostException (__FILE__,__LINE__,messy)
#define SocketException(messy)        ESocketException        (__FILE__,__LINE__,messy)
#define BindException(messy)          EBindException          (__FILE__,__LINE__,messy)
#define ListenException(messy)        EListenException        (__FILE__,__LINE__,messy)
#define ConnectException(messy)       EConnectException       (__FILE__,__LINE__,messy)
#define AcceptException(messy)        EAcceptException        (__FILE__,__LINE__,messy)
#define SocketAccessException(messy)  ESocketAccessException  (__FILE__,__LINE__,messy)

// The class

class Socket
{
 protected:

  static char* StringUDP;
  static char* StringTCP;

  struct hostent*    LocalHostInfo;   // Information on the local host
  struct hostent*    RemoteHostInfo;  // Information on the remote host
  struct sockaddr_in SocketInfo;      // The Socket Information

  port_t SocketPort;    // TCP Communication port
  int    SocketStream;  // The socket Stream
  int    AcceptStream;  // The accept Stream
  int*   IOStream;      // The current Stream
  
  index_t BufferSize;  // Transfert buffer
  
  char    HostIP[16];  // HostIp under 000x000x000x000\0 form
  bool    Verbose;     // Output of the sended/received data

 protected:
  char*   InBuffer;    // Data of this buffer
  char*   OutBuffer;   // Data of this buffer

 private:
  void printVerbose (const char* start, const char* string, const char* end);
  
 public:
           Socket(int bufferSize);
  virtual ~Socket();
 
  virtual void clientOf (const char* hostname, port_t port);
  virtual void serverOn (port_t port);

  virtual void accept() = 0;
  pid_t connection ();

  void closeSocket();
  void closeAccept();
  void closeConnection();

  virtual char readChar  () = 0;
  virtual void writeChar (char car) = 0;

  ssize_t read (char* buffer, ssize_t size);
  ssize_t write(const char* buffer, ssize_t size); 

  const char* readString();

  ssize_t writeText(const char* string);
  ssize_t writeString(const char* string, ...);
 
  ssize_t skip (ssize_t size);
 
  // informations

  virtual protocole_t getProtocole () = 0;
  virtual const char* getProtocoleName () = 0;

  virtual port_t  getPort () = 0;

  int     getDescriptor();
  index_t getBufferSize ();

  const char* getLocalHostName ();
  const char* getRemoteHostName();
  
  const char* getHostIp  ();
  const char* getHostIp  (char separator);
  
  bool getVerbose ();
  bool setVerbose (bool state);
};


#endif
