#include <signal.h>
#include <unistd.h>
#include <setjmp.h>
#include "socketUDP.h"

#define ALARM_TIME 5
#define MAX_RESEND 10

// signal handler pour les recvfrom

sigjmp_buf JumpEnv;

void signal_handler (int signalNumber)
{
  longjmp(JumpEnv,1);
}


// Constructeur / Destructeur

SocketUDP::SocketUDP(int bufferSize) : Socket(bufferSize)
{
  AcceptPort   = -1;
  SocketStream = socket(AF_INET,UDP,0);
  
  if(!SocketStream)
	{
	  closeSocket();
	  throw SocketException("Can't create Socket");
	}

	LastSended = nextValue(0);
 	WaitingFor =  nextValue(0);
}

SocketUDP::~SocketUDP()
{
}


void SocketUDP::clientOf(const char* hostname, port_t port)
{
  int returned;

  Socket::clientOf(hostname,port);

  writeString("UDP Port");
  SocketInfo.sin_port = htons(atoi(readString()));
  
  returned = ::connect(SocketStream,(sockaddr*)&SocketInfo,sizeof(SocketInfo));

  if(returned < 0)
	throw ConnectException("Can't connect to socket");

  IOStream = &SocketStream;
  
  writeString("UDP Port Ok");
}
        

void SocketUDP::accept ()
{
  int returned;
  unsigned int sizeInfo;
  char number[20];

  // Check for accept query

  if(strcmp(readString(),"UDP Port") != 0)
	throw AcceptException("Can't accept client invalid protocol");

  // Create a new Socket

  AcceptStream = socket(AF_INET,UDP,0);
  
  if(!AcceptStream)
	{
	  closeAccept();
	  throw SocketException("Can't create Socket");
	}
  
  // Give a public name to socket

  memset(&AcceptInfo,0,sizeof(AcceptInfo));

  AcceptInfo.sin_family      = LocalHostInfo->h_addrtype;
  AcceptInfo.sin_addr.s_addr = INADDR_ANY;
  AcceptInfo.sin_port        = htons(0);

  returned = bind(AcceptStream,(struct sockaddr*)&AcceptInfo,sizeof(AcceptInfo));

  if(returned)
	throw ConnectException("Can't bind to socket");

  // Get Information back on the Socket

  sizeInfo = sizeof(AcceptInfo);
  returned = getsockname(AcceptStream,(struct sockaddr*)&AcceptInfo,&sizeInfo);

  if(returned)
	throw SocketException("Can't get Socket Name");

  AcceptPort = ntohs(AcceptInfo.sin_port);

  // Send Accept port

  sprintf(number,"%d",AcceptPort);
  writeString(number);

  IOStream = &AcceptStream;

  // Read Ack
  if (strcmp("UDP Port Ok",readString()) != 0)
	throw AcceptException("Client doesnt acknowledge my port");
}


char SocketUDP::nextValue(char value)
{
	if (++value == 'A') value++;
	return value;
}


void SocketUDP::sendAck (bool ok)
{
  char ack[2];
  unsigned int sizeFrom;

  sizeFrom = sizeof(SocketInfo);
  ack[0] = 'A';
  ack[1] = ok?'Y':'N';

  sendto(getDescriptor(),(void*)ack,2,0,(struct sockaddr*)&SocketInfo,sizeFrom);
}


bool SocketUDP::recvAck ()
{
  unsigned int sizeFrom;
  char ack[2];
  
  sizeFrom = sizeof(SocketInfo);
  
  signal(SIGALRM,signal_handler);
  alarm(ALARM_TIME);
  
  if(setjmp(JumpEnv) == 0)
	{
	  recvfrom(getDescriptor(),(void*)ack,2,0,(struct sockaddr*)&SocketInfo,&sizeFrom);
	  alarm(0);

	  //printf("Recv Ack [%c][%c]\n",ack[0],ack[1]);

	  return (ack[0] == 'A' && ack[1] == 'Y');
	}
  else
	return FALSE;
}

char SocketUDP::readChar ()
{
  bool readed;
  int recvSize;
  char recvBuffer[2];
  
  unsigned int sizeFrom;
  sizeFrom = sizeof(SocketInfo);
  readed = FALSE;

  //printf(" wait "); fflush(stdout);

  while (!readed)
	{
	  signal(SIGALRM,signal_handler);
	  alarm(ALARM_TIME);

	  if(setjmp(JumpEnv) == 0)
		{	
		  recvSize = -1;
		  recvSize = recvfrom(getDescriptor(),(void*)recvBuffer,2,0,(struct sockaddr*)&SocketInfo,&sizeFrom);
		  alarm(0);
		  
		  if(recvSize != 2) // error during sending
			{
			  sendAck(FALSE);
			}
		  else if(recvBuffer[0] == 'A')
			{
			  //printf("Old ack\n");
			  // Nop Old ack ..
			}
		  else if(recvBuffer[0] != WaitingFor)  // It is not the one waited
			{
			  //printf("Recv index %d wants %d\n",recvBuffer[0],WaitingFor);
			}
		  else
			{
			  readed = TRUE;
			  sendAck(TRUE);
			  WaitingFor = nextValue (WaitingFor);
			}
		}
	  else
		{
		  //printf("Time out\n");
		  sendAck(FALSE);
		}
	}

  signal(SIGALRM,SIG_DFL);
  return recvBuffer[1];
}

void SocketUDP::writeChar(char letter)
{
  int  nbTry;
  bool writed;
  unsigned int sizeFrom;
  sizeFrom = sizeof(SocketInfo);
  nbTry = 0;

  SendBuffer[0] = LastSended;
  SendBuffer[1] = letter;

  writed = FALSE;

  //printf("      Send %c\n",letter);
  
  while(!writed)
	{
	  sendto(getDescriptor(),(void*)SendBuffer,2,0,(struct sockaddr*)&SocketInfo,sizeFrom);
	  writed = recvAck();

	  //if(0)
	  if(!writed)
		{
		  if(nbTry++ > MAX_RESEND)
			throw SocketException("Connexion problem.. The other side probably died");
		  //printf("Ack resend %c\n",letter);
		}
	}
  LastSended = nextValue(LastSended);
}


// Various

port_t SocketUDP::getPort()
{
  if (AcceptPort == -1)
	return SocketPort;
  else
	return AcceptPort;
}

protocole_t SocketUDP::getProtocole()
{
  return UDP;
}


const char* SocketUDP::getProtocoleName()
{
  return StringUDP;
}
